/*************************************************************************
 * Copyright (C) 2017-2018 Barcelona Supercomputing Center               *
 *                         Centro Nacional de Supercomputacion           *
 * All rights reserved.                                                  *
 *                                                                       *
 * This file is part of the NORNS Data Scheduler, a service that allows  *
 * other programs to start, track and manage asynchronous transfers of   *
 * data resources transfers requests between different storage backends. *
 *                                                                       *
 * See AUTHORS file in the top level directory for information           *
 * regarding developers and contributors.                                *
 *                                                                       *
 * The NORNS Data Scheduler is free software: you can redistribute it    *
 * and/or modify it under the terms of the GNU Lesser General Public     *
 * License as published by the Free Software Foundation, either          *
 * version 3 of the License, or (at your option) any later version.      *
 *                                                                       *
 * The NORNS Data Scheduler is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General             *
 * Public License along with the NORNS Data Scheduler.  If not, see      *
 * <http://www.gnu.org/licenses/>.                                       *
 *************************************************************************/

#ifndef __API_LISTENER_HPP__
#define __API_LISTENER_HPP__

#include <boost/asio.hpp>
#include <boost/filesystem.hpp>

#include "common.hpp"
#include "api/local-endpoint.hpp"
#include "api/remote-endpoint.hpp"
#include "api/signal-listener.hpp"

namespace ba = boost::asio;
namespace bfs = boost::filesystem;

namespace norns {
namespace api {

/* simple lister for an AF_UNIX socket that accepts requests asynchronously and
 * invokes a callback with a fixed-length payload */
template <typename Message>
class listener {

    using Dispatcher = dispatch_table<
        typename Message::key_type, 
        typename Message::key_hash_type, 
        std::unique_ptr<typename Message::response_type>,
        std::unique_ptr<typename Message::request_type>
            >;

    using MessageKey = typename Message::key_type;

public:
    explicit listener() :
        m_ios(),
        m_msg_handlers(std::make_shared<Dispatcher>()), 
        m_signal_listener(m_ios) { }

    ~listener() {
        Message::cleanup();
    }

    void run() {

        for(auto& endp : m_local_endpoints) {
            endp->do_accept();
        }

        for(auto& endp : m_remote_endpoints) {
            endp->do_accept();
        }

        m_signal_listener.do_accept();

        m_ios.run();
    }

    void stop() {
        m_ios.stop();
    }

    template <typename Callable>
    void register_callback(MessageKey k, Callable&& func) {
        m_msg_handlers->add(k, std::forward<Callable>(func));
    }

    /* register a socket endpoint from which to accept local requests */
    void register_endpoint(const bfs::path& sockfile) {
        m_local_endpoints.emplace_back(
                std::make_shared<local_endpoint<Message>>(sockfile, m_ios, m_msg_handlers));
    }

    /* register a socket endpoint from which to accept remote requests */
    void register_endpoint(short port) {
        m_remote_endpoints.emplace_back(
                std::make_shared<remote_endpoint<Message>>(port, m_ios, m_msg_handlers));
    }

    template <typename... Args>
    void set_signal_handler(signal_listener::SignalHandler handler, Args... signums) {
        m_signal_listener.set_handler(handler, std::forward<Args>(signums)...);
    }

    static void cleanup() {
        Message::cleanup();
    }

private:
    /*! Main io_service */
    boost::asio::io_service                m_ios;

    /*! Dispatcher of message handlers */
    std::shared_ptr<Dispatcher> m_msg_handlers;

    /*! Dispatcher of signal listener */
    signal_listener m_signal_listener;

    /*! Local endpoints */
    std::vector<std::shared_ptr<local_endpoint<Message>>>  m_local_endpoints;

    /*! Remote endpoints */
    std::vector<std::shared_ptr<remote_endpoint<Message>>> m_remote_endpoints;
};

} // namespace api
} // namespace norns

#endif /* __API_LISTENER_HPP__ */
